@@ -72,19 +72,18 @@ module Google
     end
 
     Timestamp.class_eval do
-      if RUBY_VERSION < "2.5"
-        def to_time
-          Time.at(self.to_f)
-        end
-      else
-        def to_time
-          Time.at(seconds, nanos, :nanosecond)
-        end
+      def to_time
+        Time.at(seconds, nanos, :nanosecond)
+      end
+
+      def self.from_time(time)
+        new.from_time(time)
       end
 
       def from_time(time)
         self.seconds = time.to_i
         self.nanos = time.nsec
+        self
       end
 
       def to_i
@@ -132,10 +131,14 @@ module Google
         end
       end
 
+      def self.from_ruby(value)
+        self.new.from_ruby(value)
+      end
+
       def from_ruby(value)
         case value
         when NilClass
-          self.null_value = 0
+          self.null_value = :NULL_VALUE
         when Numeric
           self.number_value = value
         when String
@@ -155,6 +158,8 @@ module Google
         else
           raise UnexpectedStructType
         end
+
+        self
       end
     end
 
@@ -225,6 +230,5 @@ module Google
         ret
       end
     end
-
   end
 end
